import game.level_management as level_management
from game.models import Level


def create_save_level(teacher_or_student, level_name="1", shared_with=None):
    data = {
        "origin": '{"coordinate":[3,5],"direction":"S"}',
        "pythonEnabled": False,
        "decor": [],
        "blocklyEnabled": True,
        "blocks": [
            {"type": "move_forwards"},
            {"type": "turn_left"},
            {"type": "turn_right"},
        ],
        "max_fuel": "50",
        "pythonViewEnabled": False,
        "character": "3",
        "name": level_name,
        "theme": 1,
        "anonymous": False,
        "cows": "[]",
        "path": '[{"coordinate":[3,5],"connectedNodes":[1]},{"coordinate":[3,4],"connectedNodes":[0]}]',
        "traffic_lights": "[]",
        "destinations": "[[3,4]]",
    }
    level = Level(default=False, anonymous=data["anonymous"])
    level.owner = teacher_or_student.user
    level_management.save_level(level, data)
    level.save()

    if shared_with is not None:
        for user in shared_with:
            level.shared_with.add(user)
        level.save()

    return level



from portal.tests.pageObjects.portal.base_page import BasePage
from selenium.webdriver.common.by import By


class EditorPage(BasePage):
    def __init__(self, browser):
        super(EditorPage, self).__init__(browser)

        assert self.on_correct_page("editor_page")

        self._dismiss_initial_dialog()

    def _dismiss_initial_dialog(self):
        self.dismiss_dialog("close_button")
        return self

    def dismiss_dialog(self, button_id):
        self.wait_for_element_to_be_clickable((By.ID, button_id), wait_seconds=15)
        self.browser.find_element(By.ID, button_id).click()
        self.wait_for_element_to_be_invisible((By.ID, button_id), wait_seconds=15)

    def go_to_code_tab(self):
        self.browser.find_element(By.ID, "blocks_tab").click()



from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [("game", "0044_auto_20150615_1156")]

    operations = [
        migrations.AddField(
            model_name="decor",
            name="z_index",
            field=models.IntegerField(default=0),
            preserve_default=False,
        )
    ]



from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from game.end_to_end_tests.base_game_test import BaseGameTest

DELAY_TIME = 10


class TestPythonLevels(BaseGameTest):
    def test_can_see_python_commands(self):
        self.run_python_commands_test(level=92)

    def test_clear_console(self):
        self.run_clear_console_test(level=92)

    def test_console_parse_error(self):
        self.run_console_parse_error_test(level=92)

    def test_console_attribute_error(self):
        self.run_console_attribute_error_test(level=92)

    def test_console_print(self):
        self.run_console_print_test(level=92)

    def test_invalid_import(self):
        self.run_invalid_import_test(level=109)

    def test_run_code(self):
        self.go_to_level(92)
        run_code = self.selenium.find_element(By.ID, "run-code-button")
        run_code.click()
        assert WebDriverWait(self.selenium, DELAY_TIME).until(
            EC.presence_of_all_elements_located((By.ID, "myModal-lead"))
        )



# Generated by Django 3.2.15 on 2022-10-31 15:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('common', '0030_add_maintenance_banner'),
        ('game', '0075_level_48_houses'),
    ]

    operations = [
        migrations.AddField(
            model_name='level',
            name='locked_for_class',
            field=models.ManyToManyField(blank=True, related_name='locked_levels', to='common.Class'),
        ),
    ]



from django.apps.registry import Apps
from django.db import migrations


def add_cows_to_existing_levels(apps: Apps, *args):
    Level = apps.get_model("game", "Level")
    Block = apps.get_model("game", "Block")
    LevelBlock = apps.get_model("game", "LevelBlock")
    LevelDecor = apps.get_model("game", "LevelDecor")

    level_38 = Level.objects.get(name="38", default=1)
    level_38.cows = '[{"minCows":1,"maxCows":1,"potentialCoordinates":[{"x":2,"y":6},{"x":3,"y":4},{"x":3,"y":1}],"type":"WHITE"}]'
    level_38.theme_name = "farm"
    level_38.model_solution = "[10]"
    level_38.save()

    level_39 = Level.objects.get(name="39", default=1)
    level_39.cows = '[{"minCows":1,"maxCows":1,"potentialCoordinates":[{"x":4,"y":4},{"x":8,"y":4},{"x":8,"y":6}],"type":"WHITE"}]'
    level_39.theme_name = "farm"
    level_39.model_solution = "[10]"
    level_39.disable_route_score = True
    level_39.save()

    level_47 = Level.objects.get(name="47", default=1)
    level_47.traffic_lights = "[]"
    level_47.cows = '[{"minCows":1,"maxCows":1,"potentialCoordinates":[{"x":6,"y":4},{"x":4,"y":6},{"x":2,"y":4}],"type":"WHITE"}]'
    level_47.theme_name = "farm"
    level_47.model_solution = "[8]"
    level_47.save()

    block_sound_horn = Block.objects.get(type="sound_horn")
    block_cow_crossing = Block.objects.get(type="cow_crossing")

    LevelBlock.objects.filter(
        level=level_47,
        type__type__in=["traffic_light", "wait"],
    ).delete()

    LevelBlock.objects.bulk_create(
        [
            LevelBlock(level=level_38, type=block_sound_horn),
            LevelBlock(level=level_38, type=block_cow_crossing),
            LevelBlock(level=level_39, type=block_sound_horn),
            LevelBlock(level=level_39, type=block_cow_crossing),
            LevelBlock(level=level_47, type=block_sound_horn),
            LevelBlock(level=level_47, type=block_cow_crossing),
        ]
    )

    LevelDecor.objects.filter(
        level_id__in=[
            level_47.id,
        ]
    ).delete()

    LevelDecor.objects.bulk_create(
        [
            LevelDecor(
                x=570,
                y=296,
                decorName="pond",
                level=level_39,
            ),
            LevelDecor(
                x=570,
                y=408,
                decorName="pond",
                level=level_39,
            ),
            LevelDecor(
                x=570,
                y=514,
                decorName="pond",
                level=level_39,
            ),
            LevelDecor(
                x=293,
                y=300,
                decorName="tree1",
                level=level_39,
            ),
            LevelDecor(
                x=711,
                y=101,
                decorName="tree1",
                level=level_39,
            ),
            LevelDecor(
                x=897,
                y=604,
                decorName="tree1",
                level=level_39,
            ),
            LevelDecor(
                x=114,
                y=315,
                decorName="tree1",
                level=level_39,
            ),
            LevelDecor(
                x=37,
                y=502,
                decorName="pond",
                level=level_47,
            ),
            LevelDecor(
                x=39,
                y=400,
                decorName="pond",
                level=level_47,
            ),
            LevelDecor(
                x=39,
                y=296,
                decorName="pond",
                level=level_47,
            ),
            LevelDecor(
                x=720,
                y=657,
                decorName="bush",
                level=level_47,
            ),
            LevelDecor(
                x=720,
                y=556,
                decorName="bush",
                level=level_47,
            ),
            LevelDecor(
                x=720,
                y=455,
                decorName="bush",
                level=level_47,
            ),
            LevelDecor(
                x=720,
                y=343,
                decorName="bush",
                level=level_47,
            ),
            LevelDecor(
                x=720,
                y=220,
                decorName="bush",
                level=level_47,
            ),
            LevelDecor(
                x=403,
                y=378,
                decorName="tree1",
                level=level_47,
            ),
            LevelDecor(
                x=440,
                y=471,
                decorName="tree1",
                level=level_47,
            ),
            LevelDecor(
                x=132,
                y=649,
                decorName="tree1",
                level=level_47,
            ),
            LevelDecor(
                x=620,
                y=689,
                decorName="tree1",
                level=level_47,
            ),
            LevelDecor(
                x=500,
                y=149,
                decorName="tree1",
                level=level_47,
            ),
            LevelDecor(
                x=550,
                y=483,
                decorName="tree2",
                level=level_47,
            ),
        ]
    )


class Migration(migrations.Migration):
    dependencies = [("game", "0082_level_43_solution")]
    operations = [
        migrations.RunPython(
            add_cows_to_existing_levels,
            reverse_code=migrations.RunPython.noop,
        )
    ]



# Generated by Django 3.2.16 on 2022-12-02 09:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('game', '0077_alter_level_next_level'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='block',
            options={'ordering': ['block_type', 'pk']},
        ),
        migrations.AddField(
            model_name='block',
            name='block_type',
            field=models.IntegerField(choices=[(0, 'Start'), (1, 'Action'), (2, 'Condition'), (3, 'Procedure'), (4, 'ControlFlow')], default=4),
            preserve_default=False,
        ),
    ]



from builtins import str

from common.models import Class, Student, UserProfile
from django.contrib.auth.models import User
from django.db import models


def theme_choices():
    from game.theme import get_all_themes

    return [(theme.name, theme.name) for theme in get_all_themes()]


def character_choices():
    from game.character import get_all_character

    return [(character.name, character.name) for character in get_all_character()]


class Block(models.Model):
    type = models.CharField(max_length=200)
    block_type = models.IntegerField(
        choices=[
            (0, "Start"),
            (1, "Action"),
            (2, "Condition"),
            (3, "Procedure"),
            (4, "ControlFlow"),
            (5, "Variable"),
            (6, "Math"),
        ]
    )

    def __str__(self):
        return self.type

    class Meta:
        ordering = ["block_type", "pk"]


class Episode(models.Model):
    """Variables prefixed with r_ signify they are parameters for random level generation"""

    name = models.CharField(max_length=200)
    next_episode = models.ForeignKey(
        "self", null=True, default=None, on_delete=models.SET_NULL
    )
    in_development = models.BooleanField(default=False)

    r_random_levels_enabled = models.BooleanField(default=False)
    r_branchiness = models.FloatField(default=0, null=True)
    r_loopiness = models.FloatField(default=0, null=True)
    r_curviness = models.FloatField(default=0, null=True)
    r_num_tiles = models.IntegerField(default=5, null=True)
    r_blocks = models.ManyToManyField(Block, related_name="episodes")
    r_blocklyEnabled = models.BooleanField(default=True)
    r_pythonEnabled = models.BooleanField(default=False)
    r_trafficLights = models.BooleanField(default=False)
    r_cows = models.BooleanField(default=False)

    @property
    def first_level(self):
        return self.levels[0]

    @property
    def levels(self):
        """Sorts the levels by integer conversion of "name" which should equate to the correct play order"""

        return sorted(self.level_set.all(), key=lambda level: int(level.name))

    @property
    def difficulty(self):
        """
        Maps the Episode's id to a difficulty level, used later on to map to CSS classes and different background
        colours
        """
        difficulty_map = {
            1: "easy",
            2: "easy",
            3: "easy",
            4: "easy",
            5: "medium",
            6: "medium",
            7: "medium-hard",
            8: "medium-hard",
            9: "brainteasers",
            10: "hard",
            11: "advanced",
            12: "loops",
            13: "loops",
            14: "loops",
            15: "loops",
        }

        return difficulty_map.get(self.id, "easy")

    def __str__(self):
        return f"Episode: {self.name}"


class LevelManager(models.Manager):
    def sorted_levels(self):
        # Sorts all the levels by integer conversion of "name" which should equate to the correct play order
        # Custom levels do not have an episode

        return sort_levels(
            self.model.objects.filter(episode__isnull=False).exclude(
                episode__name__icontains="coming soon"
            )
        )


def sort_levels(levels):
    return sorted(levels, key=lambda level: int(level.name))


class Level(models.Model):
    name = models.CharField(max_length=100)
    episode = models.ForeignKey(
        Episode, blank=True, null=True, default=None, on_delete=models.PROTECT
    )
    path = models.TextField(max_length=10000)
    traffic_lights = models.TextField(max_length=10000, default="[]")
    cows = models.TextField(max_length=10000, default="[]")
    origin = models.CharField(max_length=50, default="[]")
    destinations = models.CharField(max_length=50, default="[[]]")
    default = models.BooleanField(default=False)
    owner = models.ForeignKey(
        UserProfile,
        related_name="levels",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
    )
    fuel_gauge = models.BooleanField(default=True)
    max_fuel = models.IntegerField(default=50)
    direct_drive = models.BooleanField(default=False)
    next_level = models.ForeignKey(
        "self",
        null=True,
        default=None,
        on_delete=models.SET_NULL,
        related_name="prev_level",
    )
    shared_with = models.ManyToManyField(User, related_name="shared", blank=True)
    model_solution = models.CharField(blank=True, max_length=20, default="[]")
    disable_route_score = models.BooleanField(default=False)
    disable_algorithm_score = models.BooleanField(default=False)
    threads = models.IntegerField(blank=False, default=1)
    blocklyEnabled = models.BooleanField(default=True)
    pythonEnabled = models.BooleanField(default=True)
    pythonViewEnabled = models.BooleanField(default=False)
    theme_name = models.CharField(
        max_length=10, choices=theme_choices(), blank=True, null=True, default=None
    )
    character_name = models.CharField(
        max_length=20, choices=character_choices(), blank=True, null=True, default=None
    )
    anonymous = models.BooleanField(default=False)
    locked_for_class = models.ManyToManyField(
        Class, blank=True, related_name="locked_levels"
    )
    objects = LevelManager()

    def __str__(self):
        return f"Level {self.name}"

    @property
    def theme(self):
        from game.theme import get_theme

        try:
            return get_theme(self.theme_name)
        except KeyError:
            return None

    @theme.setter
    def theme(self, val):
        from game.theme import get_theme_by_pk

        self.theme_name = get_theme_by_pk(val.pk).name

    @property
    def character(self):
        from game.character import get_character

        try:
            return get_character(self.character_name)
        except KeyError:
            return None

    @character.setter
    def character(self, val):
        from game.character import get_character_by_pk

        self.character_name = get_character_by_pk(val.pk).name

    @property
    def difficulty(self):
        return self.episode.difficulty if self.episode else "easy"


class LevelBlock(models.Model):
    type = models.ForeignKey(Block, on_delete=models.CASCADE)
    level = models.ForeignKey(Level, on_delete=models.CASCADE)
    number = models.PositiveIntegerField(default=None, null=True)


class LevelDecor(models.Model):
    x = models.IntegerField()
    y = models.IntegerField()
    level = models.ForeignKey(Level, on_delete=models.CASCADE)
    decorName = models.CharField(max_length=100, default="tree1")


class Workspace(models.Model):
    name = models.CharField(max_length=200)
    owner = models.ForeignKey(
        UserProfile,
        related_name="workspaces",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
    )
    contents = models.TextField(default="")
    python_contents = models.TextField(default="")
    blockly_enabled = models.BooleanField(default=False)
    python_enabled = models.BooleanField(default=False)

    def __str__(self):
        return str(self.name)


class Attempt(models.Model):
    start_time = models.DateTimeField(auto_now_add=True)
    level = models.ForeignKey(Level, related_name="attempts", on_delete=models.CASCADE)
    student = models.ForeignKey(
        Student,
        related_name="attempts",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
    )
    finish_time = models.DateTimeField(null=True, blank=True)
    score = models.FloatField(default=0, null=True)
    workspace = models.TextField(default="")
    night_mode = models.BooleanField(default=False)
    python_workspace = models.TextField(default="")
    is_best_attempt = models.BooleanField(db_index=True, default=False)

    def elapsed_time(self):
        return self.finish_time - self.start_time



from common.tests.utils.classes import create_class_directly
from common.tests.utils.organisation import create_organisation_directly
from common.tests.utils.student import create_school_student_directly
from common.tests.utils.teacher import signup_teacher_directly
from hamcrest import assert_that, ends_with, equal_to
from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.common.by import By

from game.end_to_end_tests.base_game_test import BaseGameTest
from game.models import Attempt, Episode


class TestLevelSelection(BaseGameTest):
    def test_coins(self):

        # Set up student
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        klass, name, access_code = create_class_directly(email)
        name, password, student = create_school_student_directly(access_code)

        # Student has perfect score for each level in an episode
        episode = Episode.objects.get(name="Loops with Conditions")
        for a_level in episode.levels:
            attempt = Attempt(student=student, score=20.0, level=a_level)
            attempt.save()

        #  Student logs in
        code_page = self.go_to_homepage().go_to_student_login_page()
        student_login_page = code_page.student_input_access_code(access_code)
        page = student_login_page.student_login(name, password)

        # Goes to rapid router levels
        page = self.go_to_reverse("levels")

        # The coin images for the levels
        level_coin_images = page.browser.find_elements(By.CSS_SELECTOR, ("#collapse-4 div img"))
        # There are 4 levels in this episode, each with gold coin
        assert_that(len(level_coin_images), equal_to(4))

        for image in level_coin_images:
            assert_that(
                image.get_attribute("src"),
                ends_with("/static/game/image/coins/coin_gold.svg"),
            )

        # So the episode has a gold coin too
        episode_coin_image = page.browser.find_element(By.CSS_SELECTOR, "#episode-4 > p > img")
        assert_that(
            episode_coin_image.get_attribute("src"),
            ends_with("/static/game/image/coins/coin_gold.svg"),
        )

        try:
            image_for_uncomplete_episode = page.browser.find_element(By.CSS_SELECTOR, "#episode-3 > p > img")
        except NoSuchElementException as this_should_happen:
            pass
